#include "emlabcpp/static_vector.h"
#include "fw/servo/holder.h"

#pragma once

namespace em = emlabcpp;

namespace schpin
{

template < std::size_t N >
class servo_group
{

        // TODO: instead of optional, make servo_holder nullable
        std::array< std::optional< servo_holder >, N > servos_;

        std::size_t poll_i_ = 0;

      public:
        bool any_has_task() const
        {
                return any_has_task( servo_flags{}.set() );
        }
        bool any_has_task( servo_flags filter ) const
        {
                return em::any_of( servos_, [&]( const std::optional< servo_holder >& s ) {
                        return s && filter[s->get_id()] && s->has_task();
                } );
        }

        template < typename Visitor >
        auto make_matcher( Visitor&& vis )
        {
                return match(
                    [&]( control_tag, servo_id id, servo_master_message m ) {
                            vis( control_tag{}, fw_node_group::create< SERVO >( id, m ) );
                    },
                    [&]( servo_tag, s_ctl_id msg_id, reply r, auto m ) {
                            vis( servo_tag{}, msg_id, r, m );
                    },
                    [&]( activate_tag, servo_id id ) {
                            vis( control_tag{},
                                 fw_node_group::create< ACTIVATE >( id, ACTIVATED ) );
                    } );
        }

        template < typename Visitor >
        void on_servo_receive( clk_time                    clock_time,
                               s_ctl_id                    s_id,
                               const lewan_master_message& m,
                               Visitor&&                   vis )
        {
                lewan_master_protocol::extract( m )
                    .bind_left( em::uncurry |
                                [&]( lewan_servo_id id, lewan_master_submessage m ) {
                                        return lewan_master_group::extract( m ).convert_left(
                                            [&]( auto var ) {
                                                    return std::make_tuple( id, var );
                                            } );
                                } )
                    .match(
                        em::uncurry |
                            [&]( lewan_servo_id id, lewan_master_variant msg_var ) {
                                    auto iter = find_servo( *id );

                                    if ( iter == servos_.end() ) {
                                            reply_ctl_error< SERVO_ID_MISSING_E >(
                                                ctl_id{ 0 }, vis, *id );
                                            return;
                                    }
                                    auto opt_msg =
                                        iter->value().on_servo_receive( clock_time, s_id, msg_var );
                                    if ( opt_msg ) {
                                            vis( control_tag{},
                                                 fw_node_group::create< SERVO >( *id, *opt_msg ) );
                                    }
                            },
                        [&]( auto e ) {  // TODO: add error handling
                                SCHPIN_ERROR_LOG( "lewan.group", "Lewan error: " << e );
                        } );
        }

        template < typename Visitor >
        void on_servo_not_replying( const master_lewan_message& m, Visitor&& vis )
        {
                master_lewan_protocol::match(
                    m,
                    [&]( error_tag, auto... ) {  // TODO: add error handling

                    },
                    [&]( lewan_servo_id sid, master_lewan_submessage ) {
                            auto iter = find_servo( *sid );

                            if ( iter->value().is_activating() ) {
                                    iter->reset();

                                    reply_ctl< ACTIVATE >( vis, *sid, MISSING );
                            }
                    } );
        }

        template < typename Visitor >
        void tick( clk_time clock_time, Visitor&& vis )
        {
                for ( std::optional< servo_holder >& s : servos_ ) {
                        if ( !s ) {
                                continue;
                        }
                        s->tick( clock_time, make_matcher( vis ) );
                }

                polling_tick( clock_time, vis );
        }

        template < typename Visitor >
        void stop( Visitor&& v )
        {
                for ( std::optional< servo_holder >& s : servos_ ) {
                        s->stop( make_matcher( v ) );
                }
        }

        template < typename Visitor >
        void activate( clk_time now, ctl_id cid, servo_addr saddr, Visitor&& v )
        {
                auto [stype, sid] = saddr;
                SCHPIN_ERROR_LOG( "group", "Activating: " << sid );

                auto iter = find_servo( sid );
                if ( iter != servos_.end() ) {
                        if ( iter->value().is_activating() ) {
                                reply_ctl< ACTIVATE >( v, sid, WAS_ACTIVATING );
                        } else {
                                reply_ctl< ACTIVATE >( v, sid, WAS_ACTIVATED );
                        }
                        SCHPIN_ERROR_LOG( "group", "Allready got it" );
                        return;
                }

                bool is_full = em::all_of( servos_ );

                if ( is_full ) {
                        SCHPIN_ERROR_LOG( "group", "Is full" );
                        reply_ctl_error< SERVO_CAPACITY_USED_E >( cid, v );
                        return;
                }

                for ( std::optional< servo_holder >& spot : servos_ ) {
                        if ( !spot ) {
                                SCHPIN_ERROR_LOG( "group", "Placed new servo" );
                                spot = servo_holder::make( stype, sid, now );
                                break;
                        }
                }

                reply_ctl< ACTIVATE >( v, sid, ACTIVATING );
        }

        template < typename Visitor >
        void deactivate( ctl_id, servo_id sid, Visitor&& vis )
        {
                SCHPIN_ERROR_LOG( "group", "Deactivating: " << sid );
                auto iter = find_servo( sid );
                if ( iter == servos_.end() ) {
                        reply_ctl_error< SERVO_ID_MISSING_E >( ctl_id{ 0 }, vis, sid );
                        return;
                }
                iter->value().stop( make_matcher( vis ) );
                iter->reset();
        }

        template < typename Visitor >
        void on_event( ctl_id id, clk_time clock_time, servo_event e, Visitor&& vis )
        {
                auto iter = find_servo( e.id );
                SCHPIN_ASSERT( iter != servos_.end() );
                // TODO: check!
                iter->value().on_event( id, clock_time, e.sub_event, make_matcher( vis ) );
        }

        em::either< servo_event, fw_node_group_message >
        on_servo_packet( ctl_id id, servo_id sid, master_servo_command_variant subvar )
        {
                auto iter = find_servo( sid );
                if ( iter == servos_.end() ) {
                        return { make_ctl_error< SERVO_NOT_ACTIVE_E >( id, sid ) };
                }

                return iter->value()
                    .on_servo_packet( id, subvar )
                    .convert_right( [&]( servo_master_message m ) {
                            return fw_node_group::create< SERVO >( sid, m );
                    } );
        }

      private:
        template < typename Visitor >
        void polling_tick( clk_time ctime, Visitor&& vis )
        {
                if ( em::none_of( servos_ ) ) {
                        return;
                }
                do {
                        poll_i_ = ( poll_i_ + 1 ) % servos_.size();
                } while ( !servos_[poll_i_] );

                servos_[poll_i_]->poll( ctime, poll_i_, make_matcher( vis ) );
        }

        auto find_servo( servo_id id )
        {
                return em::find_if( servos_, [&]( const std::optional< servo_holder >& j ) {
                        return j && j->get_id() == id;
                } );
        }
};

}  // namespace schpin
