#include "fw/dynamixel/servo.h"
#include "fw/lewan/servo.h"
#include "fw/lewan/tester.h"

#pragma once

namespace schpin
{

using servo_events_variant = std::variant< lewan_servo_events_variant, dynmx_servo_events_variant >;

struct servo_event
{
        servo_id             id;
        servo_events_variant sub_event;
};

class servo_holder
{
        // TODO: using std::variant+std::visit may not be the best idea ever, just fix this to
        // inheritance/interface
        using servos_var = std::variant< lewan_servo, dynmx_servo, lewan_servo_tester >;

        servos_var servo_;

        servo_holder( servos_var var )
          : servo_( std::move( var ) )
        {
        }

        public:

        static std::optional< servo_holder > make( servo_type st, servo_id sid, clk_time now )
        {
                auto opt_lid = lewan_servo_id::make( sid );
                auto opt_did = dynmx_servo_id::make( sid );
                switch ( st ) {
                        case LEWAN:
                                if ( !opt_lid ) {
                                        return {};
                                }
                                return { servo_holder{ lewan_servo{ *opt_lid, now } } };
                        case LEWAN_TEST:
                                if ( !opt_lid ) {
                                        return {};
                                }
                                return { servo_holder{ lewan_servo_tester{ *opt_lid, now } } };
                        case DYNMX:
                                if ( !opt_did ) {
                                        return {};
                                }
                                return { servo_holder{ dynmx_servo{ *opt_did } } };
                }
                return {};
        }

        template < typename T >
        std::optional< servo_master_message > on_servo_receive( clk_time clock_time,
                                                                s_ctl_id s_id,
                                                                const T& var )
        {
                return std::visit(
                    [&]( auto& servo ) -> std::optional< servo_master_message > {
                            using receive_type =
                                typename std::decay_t< decltype( servo ) >::receive_type;
                            if constexpr ( std::is_same_v< std::decay_t< T >, receive_type > ) {
                                    return servo.on_servo_receive( clock_time, s_id, var );
                            } else {
                                    // TODO: report error
                                    return {};
                            }
                    },
                    servo_ );
        }

        servo_id get_id() const
        {
                return std::visit(
                    [&]( const auto& servo ) {
                            return servo.get_id();
                    },
                    servo_ );
        }

        bool is_activating() const
        {
                return std::visit(
                    [&]( const auto& servo ) {
                            return servo.is_activating();
                    },
                    servo_ );
        }

        bool has_task() const
        {
                return std::visit(
                    [&]( const auto& servo ) {
                            return servo.has_task();
                    },
                    servo_ );
        }

        template < typename Visitor >
        void stop( Visitor&& vis )
        {
                std::visit(
                    [&]( auto& servo ) {
                            servo.stop( vis );
                    },
                    servo_ );
        }

        template < typename Visitor >
        void tick( clk_time clock_time, Visitor&& vis )
        {
                std::visit(
                    [&]( auto& servo ) {
                            servo.tick( clock_time, vis );
                    },
                    servo_ );
        }

        em::either< servo_event, servo_master_message >
        on_servo_packet( ctl_id id, master_servo_command_variant subcmd )
        {
                return std::visit(
                    [&]( auto& servo ) {
                            return servo.on_servo_packet( id, subcmd )
                                .convert_left( [&]( auto event ) {
                                        return servo_event{ servo.get_id(), event };
                                } );
                    },
                    servo_ );
        }

        template < typename Visitor >
        void on_event( ctl_id cid, clk_time clock_time, servo_events_variant var, Visitor&& vis )
        {
                std::visit(
                    [&]( auto& servo ) {
                            using events_variant =
                                typename std::decay_t< decltype( servo ) >::events_variant;
                            if ( std::holds_alternative< events_variant >( var ) ) {
                                    servo.on_event(
                                        cid, clock_time, std::get< events_variant >( var ), vis );
                            } else {
                                    // TODO: REPORT ERROR
                                    return;
                            }
                    },
                    servo_ );
        }

        template < typename Visitor >
        void poll( clk_time ctime, std::size_t i_offset, Visitor&& vis )
        {
                std::visit(
                    [&]( auto& servo ) {
                            servo.poll( ctime, i_offset, vis );
                    },
                    servo_ );
        }

        void dump_info()
        {
                std::visit(
                    [&]( auto& servo ) {
                            servo.dump_info();
                    },
                    servo_ );
        }
};

}  // namespace schpin
