#include "node/base.h"
#include "node/config.h"
#include "shared/protocol.h"

#pragma once

namespace schpin
{

class servo
{
        std::string                        name_;
        std::string                        joint_name_;
        servo_id                           id_;
        prot_config_loader< servo_config > s_config_;
        em::angle                          state_;

      public:
        servo( std::string name, const node_servo_config& c )
          : name_( std::move( name ) )
          , joint_name_( c.joint_name )
          , id_( c.id )
        {
                s_config_->input_low  = bangle::from_float( *c.from_angle );
                s_config_->input_high = bangle::from_float( *c.to_angle );
                s_config_->servo_low  = c.from_servo;
                s_config_->servo_high = c.to_servo;
        }

        const servo_config* operator->() const
        {
                return &*s_config_;
        }

        const std::string& get_name() const
        {
                return name_;
        }

        const std::string& get_joint_name() const
        {
                return joint_name_;
        }

        const servo_id& get_id() const
        {
                return id_;
        }

        template < typename Visitor >
        void set_config( const nlohmann::json& j, Visitor&& vis )
        {
                from_json( j, *s_config_ );
                SCHPIN_INFO_LOG( "wololo", "setting config to: " << nlohmann::json{ *s_config_ } );
                s_config_.load_config( [&]( master_servo_config_message m ) {
                        vis( make_message< CONFIG >( m ) );
                } );
        }

        node_fw_group_message make_pos_message() const
        {
                return make_message< POSITION >();
        }

        node_fw_group_message make_goto_message( bangle a, btime t ) const
        {
                return make_message< GOTO >( a, t );
        }

        node_fw_group_message make_spin_message( bangular_velocity av, btime t ) const
        {
                return make_message< SPIN >( av, t );
        }

        node_fw_group_message make_spin_until_message( bangular_velocity av, bangle a ) const
        {
                return make_message< SPIN_UNTIL >( a, av );
        }

        node_fw_group_message make_activate_message() const
        {
                // TODO: this actually may be an error
                return node_fw_group::create< ACTIVATE >( servo_addr{ LEWAN, id_ } );
        }

        template < typename Visitor >
        void on_feedback( servo_master_command_variant v, Visitor&& vis )
        {
                // this is not a good approach, the "boot" of servos is not well defined in APP,
                // there should be proper process for it -> servo is actiavted, than config is
                // loaded
                if ( !s_config_.has_loaded() ) {
                        s_config_.load_config( [&]( master_servo_config_message m ) {
                                vis( make_message< CONFIG >( m ) );
                        } );
                }

                v << tuple_match(  //
                    [&]( cmd_tag< POSITION >, bangle a ) {
                            state_ = em::angle{ a.as_float() };
                            vis( *this, state_ );
                    },
                    [&]( cmd_tag< CONFIG >, servo_master_config_variant var ) {
                            s_config_.on_servo_config_message(
                                var, [&]( master_servo_config_message m ) {
                                        vis( make_message< CONFIG >( m ) );
                                } );
                    },
                    [&]( auto... ) {} );
        }

      private:
        template < messages_enum CMD, typename... Args >
        node_fw_group_message make_message( Args... args ) const
        {
                return { node_fw_group::create< SERVO >(
                    id_, master_servo_group::create< CMD >( args... ) ) };
        }
};

inline std::ostream& operator<<( std::ostream& os, const servo& s )
{
        return os << s.get_name() << " - " << s.get_joint_name() << " - " << s.get_id() << ","
                  << s->input_low << "-" << s->input_high << "," << s->servo_low << "-"
                  << s->servo_high << " - " << s->angle_change_report << " - "
                  << s->angle_stall_range;
}

class servo_group
{
        std::vector< servo > servos_;

      public:
        servo_group( const std::map< std::string, node_servo_config >& configs )
          : servos_(
                map_f_to_v( configs, [&]( const std::pair< std::string, node_servo_config >& cp ) {
                        return servo{ cp.first, cp.second };
                } ) )
        {
        }

        const std::vector< servo >& get_servos() const
        {
                return servos_;
        }

        auto find_servo( servo_id id )
        {
                return em::find_if( servos_, [&]( const servo& s ) {  //
                        return s.get_id() == id;
                } );
        }

        template < typename Visitor >
        void set_servo_config( servo_id id, const nlohmann::json& j, Visitor&& vis )
        {
                auto iter = find_servo( id );
                if ( iter == servos_.end() ) {
                        SCHPIN_ERROR_LOG(
                            "servos", "Got servo config command for non-existing servo: " << id );
                        return;
                }
                iter->set_config( j, vis );
        }

        template < typename Visitor >
        void on_feedback( servo_id id, servo_master_command_variant v, Visitor&& vis )
        {
                auto iter = find_servo( id );
                if ( iter == servos_.end() ) {
                        SCHPIN_ERROR_LOG( "servos",
                                          "Got servo command for non-existing servo: " << id );
                        return;
                }

                iter->on_feedback( v, vis );
        }
};

inline std::ostream& operator<<( std::ostream& os, const servo_group& g )
{
        os << "servo group: " << std::endl;
        for ( const servo& s : g.get_servos() ) {
                os << s << std::endl;
        }
        return os;
}

}  // namespace schpin
