#include "fw/base.h"
#include "fw/config.h"
#include "fw/lewan/config.h"
#include "fw/lewan/protocol.h"

#include <schpin_lib_exp/protocol.h>
#include <schpin_lib_exp/scaled.h>

#pragma once

namespace schpin
{
enum messages_enum : uint8_t
{
        PING                = 1,
        PONG                = 2,
        GOTO                = 3,
        STOP                = 4,
        SYNC                = 5,
        SPIN                = 6,
        CMD_HASH            = 7,
        TELEM               = 8,
        POSITION            = 9,
        ERROR               = 10,
        ACTIVATE            = 11,
        TEMP                = 12,
        VOLTAGE             = 13,
        VELOCITY            = 14,
        SPIN_UNTIL          = 15,
        CONFIG              = 16,
        SERVO               = 18,
        FINISHED            = 19,
        INPUT_RANGE         = 20,
        SERVO_RANGE         = 21,
        ANGLE_CHANGE_REPORT = 22,
        INPUT_BUFFER_SIZE   = 23,
        EVENT_BUFFER_SIZE   = 24,
        TASK                = 25,
        DEACTIVATE          = 26,
};

enum error_protocol_enum : uint8_t
{
        SERVO_NOT_ACTIVE_E             = 1,
        INPUT_EVENT_BUFFER_FULL_E      = 2,
        SERVO_CAPACITY_USED_E          = 3,
        SERVO_MSG_UNKNOWN_E            = 4,
        INPUT_CONTROL_BUFFER_FULL_E    = 5,
        INPUT_CONTROL_INVALID_SIZE_E   = 6,
        INPUT_CONTROL_UNKNOWN_E        = 7,
        INPUT_INVALID_VALUE_E          = 8,
        SERVO_FEEDBACK_INVALID_SIZE_E  = 9,
        SERVO_FEEDBACK_UNKNOWN_E       = 10,
        SERVO_ID_MISSING_E             = 11,
        SERVO_UNDEFINED_POSITION_E     = 12,
        SERVO_FEEDBACK_INVALID_VALUE_E = 13,
        SERVO_NOT_RESPONDING           = 14,
        EVENT_BUFFER_NOT_EMPTY_E       = 15,
};

// TODO: revision of errors
struct error_group
  : command_group<>::with_commands<
        command< SERVO_FEEDBACK_INVALID_VALUE_E >::with_args< lewan_message_enum, uint32_t >,
        command< INPUT_EVENT_BUFFER_FULL_E >,
        command< SERVO_NOT_ACTIVE_E >::with_args< servo_id >,
        command< SERVO_CAPACITY_USED_E >,
        command< SERVO_MSG_UNKNOWN_E >::with_args< lewan_master_message >,
        command< INPUT_CONTROL_BUFFER_FULL_E >,
        command< INPUT_CONTROL_INVALID_SIZE_E >,
        command< INPUT_CONTROL_UNKNOWN_E >,
        command< INPUT_INVALID_VALUE_E >::with_args< argument_index >,
        command< SERVO_ID_MISSING_E >::with_args< servo_id >,
        command< EVENT_BUFFER_NOT_EMPTY_E > >
{
};

using error_message = typename error_group::message_t;
using error_variant = typename error_group::command_variant;

struct servo_error_group
  : command_group<>::with_commands<
        command< SERVO_FEEDBACK_UNKNOWN_E >,
        command< INPUT_CONTROL_INVALID_SIZE_E >,
        command< SERVO_FEEDBACK_INVALID_SIZE_E >,
        command< INPUT_INVALID_VALUE_E >::with_args< messages_enum, uint32_t >,
        command< SERVO_UNDEFINED_POSITION_E > >
{
};

using servo_error_message = typename servo_error_group::message_t;
using servo_error_variant = typename servo_error_group::command_variant;

struct master_servo_group
  : command_group<>::with_commands<
        command< GOTO >::with_args< bangle, btime >,
        command< SPIN >::with_args< bangular_velocity, btime >,
        command< SPIN_UNTIL >::with_args< bangle, bangular_velocity >,
        command< POSITION >,
        command< TEMP >,
        command< VOLTAGE >,
        command< VELOCITY >,
        command< TASK >,
        command< CONFIG >::with_promoted_args< master_servo_config_message > >
{
};

using master_servo_message         = typename master_servo_group::message_t;
using master_servo_command_variant = typename master_servo_group::command_variant;

struct servo_master_group
  : command_group<>::with_commands<
        command< POSITION >::with_args< bangle >,
        command< TEMP >::with_args< btemp >,
        command< VOLTAGE >::with_args< bvoltage >,
        command< VELOCITY >::with_args< bangular_velocity >,
        command< TASK >::with_args< servo_task_enum >,
        command< CONFIG >::with_promoted_args< servo_master_config_message >,
        command< ERROR >::with_args< ctl_id >::with_promoted_args< servo_error_message >,
        command< TELEM >::with_args< bangle, bangle, bangular_velocity > >
{
};

using servo_master_message         = typename servo_master_group::message_t;
using servo_master_command_variant = typename servo_master_group::command_variant;

struct node_fw_group
  : command_group<>::with_commands<
        command< PING >,
        command< STOP >,
        command< ACTIVATE >::with_args< servo_addr >,
        command< DEACTIVATE >::with_args< servo_id >,
        command< SYNC >::with_args< servo_flags >,
        command< SERVO >::with_args< servo_id >::with_promoted_args< master_servo_message >,
        command< CONFIG >::with_promoted_args< node_fw_config_message > >
{
};

using node_fw_group_message = typename node_fw_group::message_t;
using node_fw_group_variant = typename node_fw_group::command_variant;

enum activated_enum : uint8_t
{
        ACTIVATED      = 0,
        ACTIVATING     = 1,
        WAS_ACTIVATING = 2,
        WAS_ACTIVATED  = 3,
        MISSING        = 4,
};

struct fw_node_group
  : command_group<>::with_commands<
        command< PONG >,
        // command< CMD_HASH >::with_args< commandHash >,
        command< ACTIVATE >::with_args< servo_id, activated_enum >,
        command< FINISHED >::with_args< ctl_id >,
        command< INPUT_BUFFER_SIZE >::with_args< uint16_t >,
        command< EVENT_BUFFER_SIZE >::with_args< uint16_t >,
        command< ERROR >::with_args< ctl_id >::with_promoted_args< error_message >,
        command< SERVO >::with_args< servo_id >::with_promoted_args< servo_master_message >,
        command< CONFIG >::with_promoted_args< fw_node_config_message > >
{
};

using fw_node_group_message = typename fw_node_group::message_t;
using fw_node_group_variant = typename fw_node_group::command_variant;

template < typename Submsg >
using control_protocol = protocol_tuple<>::with_items<  //
    ctl_id,
    protocol_recursive_promoter< Submsg >,
    checksum >;

using fw_node_protocol = control_protocol< fw_node_group_message >;
using fw_node_message  = typename fw_node_protocol::message_t;
using fw_node_tuple    = typename fw_node_protocol::output_type;

using node_fw_protocol = control_protocol< node_fw_group_message >;
using node_fw_message  = typename node_fw_protocol::message_t;
using node_fw_tuple    = typename node_fw_protocol::output_type;

em::either< node_fw_tuple, protocol_error_record > parse_node_fw( node_fw_message );
em::either< fw_node_tuple, protocol_error_record > parse_fw_node( fw_node_message );

template < messages_enum ID, typename Visitor, typename... Args >
void reply_ctl( Visitor&& v, Args&&... args )
{
        v( control_tag{}, fw_node_group::create< ID >( args... ) );
}

template < error_protocol_enum ID, typename Visitor, typename... Args >
void reply_ctl_error( ctl_id cid, Visitor&& v, Args&&... args )
{
        reply_ctl< ERROR >( v, cid, error_group::create< ID >( args... ) );
}

template < messages_enum ID, typename... Args >
fw_node_group_message make_ctl( Args&&... args )
{
        return fw_node_group::create< ID >( args... );
}

template < error_protocol_enum ID, typename... Args >
fw_node_group_message make_ctl_error( ctl_id id, Args&&... args )
{
        return make_ctl< ERROR >( id, error_group::create< ID >( args... ) );
}

}  // namespace schpin
