#include "fw/servo/group.h"
#include "schpin_lib/logging.h"
#include "test/servo.h"

#include <gtest/gtest.h>
#include <list>
#include <random>

using namespace schpin;

std::vector< servo_config > generate_servo_params()
{
        std::vector< float > input_points = { 0.f, 0.01f, 1.f, 1.5f, 3.141592f, 10.f };

        std::vector< std::pair< bangle, bangle > > input_ranges;
        for ( auto [from, to] : combine( input_points, input_points ) ) {
                if ( abs( from - to ) < 0.5f ) {
                        continue;
                }
                input_ranges.emplace_back( bangle::from_float( from ), bangle::from_float( to ) );
        }

        std::vector< lewan_angle > servo_points = { lewan_angle::get< 0 >(),
                                                    lewan_angle::get< 10 >(),
                                                    lewan_angle::get< 500 >(),
                                                    lewan_angle::get< 980 >(),
                                                    lewan_angle::get< 1000 >() };
        std::vector< std::pair< lewan_angle, lewan_angle > > servo_ranges;
        for ( auto [sfrom, sto] : combine( servo_points, servo_points ) ) {
                if ( abs( *sfrom - *sto ) < 50 ) {
                        continue;
                }
                servo_ranges.emplace_back( sfrom, sto );
        }

        std::vector< lewan_angle > change_reports = { lewan_angle::get< 5 >(),
                                                      lewan_angle::get< 25 >() };
        std::vector< lewan_angle > stall_ranges   = { lewan_angle::get< 5 >(),
                                                    lewan_angle::get< 25 >() };

        auto raw_tests = map_f_to_v(
            combine( input_ranges, servo_ranges, change_reports, stall_ranges ),
            em::uncurry | [&]( auto ipair, auto spair, auto change_report, auto stall_range ) {
                    return servo_config{ ipair.first,
                                         ipair.second,
                                         spair.first,
                                         spair.second,
                                         change_report,
                                         stall_range,
                                         pid_gain::from_float( 3.f ),
                                         pid_gain::from_float( 0.2f ),
                                         pid_gain::from_float( 0.02f ) };
            } );
        std::vector< servo_config > res;
        std::sample(
            raw_tests.begin(), raw_tests.end(), std::back_inserter( res ), 5, std::mt19937{ 42 } );
        return res;
}

class servo_text_fixture : public ::testing::TestWithParam< servo_config >
{
      protected:
        servo_holder                       servo;
        clk_time                           clock_time;
        prot_config_loader< servo_config > config;

        servo_text_fixture()
          : servo( *servo_holder::make( LEWAN_TEST, 5, this->clock_time ) )
        {
                SetUp();
        }

        void SetUp() override
        {
                servo      = *servo_holder::make( LEWAN_TEST, 5, this->clock_time );
                clock_time = clk_time{ 0 };
                config     = prot_config_loader< servo_config >( GetParam() );
                config.load_config( [&]( master_servo_config_message msg ) {
                        on_config_message( msg );
                } );
        }

        auto make_matcher()
        {
                return match(
                    [this]( control_tag, servo_id, servo_master_message m ) {  //
                            on_servo_ctl_message( m );
                    },
                    [this]( servo_tag, s_ctl_id, reply, const auto& ) {  //
                            FAIL();
                    },
                    []( activate_tag, servo_id ) {} );
        }

        void on_servo_ctl_message( servo_master_message m )
        {
                servo_master_group::extract( m ).match(
                    [&]( servo_master_command_variant var ) {
                            var << tuple_match(
                                [&]( cmd_tag< CONFIG >, servo_master_config_variant cvar ) {
                                        config.on_servo_config_message(
                                            cvar, [&]( master_servo_config_message msg ) {
                                                    on_config_message( msg );
                                            } );
                                },
                                [&]( cmd_tag< POSITION >, bangle ) {},
                                [&]( cmd_tag< TELEM >, auto... ) {},
                                [&]( auto cmd, auto... ) {
                                        FAIL() << "with arg: " << cmd;
                                } );
                    },
                    [&]( auto... ) {
                            FAIL();
                    } );
        }

        void on_config_message( master_servo_config_message msg )
        {
                auto full_msg = master_servo_group::create< CONFIG >( msg );

                master_servo_group::extract( full_msg )
                    .match(
                        [&]( master_servo_command_variant cmd ) {
                                servo.on_servo_packet( ctl_id{ 0 }, cmd )
                                    .match(
                                        [&]( servo_event e ) {
                                                on_config_event( e );
                                        },
                                        [&]( servo_master_message ) {
                                                FAIL();
                                        } );
                        },
                        [&]( auto... ) {
                                FAIL();
                        } );
        }

        void on_config_event( servo_event e )
        {
                servo.on_event( ctl_id{ 0 }, this->clock_time, e.sub_event, make_matcher() );
        }

        void tick_n( std::size_t i, bool read_pos = false, clk_time step = clk_time{ 1 } )
        {
                for ( std::size_t j = 0; j < i; j++ ) {
                        this->clock_time += step;
                        servo.tick( this->clock_time, this->make_matcher() );
                        if ( read_pos ) {
                                servo.poll( this->clock_time, 0, this->make_matcher() );
                        }
                }
        }

        template < auto CMD, typename... Args >
        void event_cmd( Args... args )
        {
                servo
                    .on_servo_packet(
                        ctl_id{},
                        std::tuple< cmd_tag< CMD >, Args... >( cmd_tag< CMD >{}, args... ) )
                    .match(
                        [&]( servo_event event ) {
                                servo.on_event(
                                    ctl_id{}, this->clock_time, event.sub_event, make_matcher() );
                        },
                        [&]( servo_master_message m ) {
                                FAIL() << m.stringify();
                        } );
        }

        bangle get_pos()
        {
                bangle res;
                servo.on_servo_packet( ctl_id{}, std::tuple< cmd_tag< POSITION > >() )
                    .match(
                        [&]( servo_event ) {
                                FAIL();
                        },
                        [&]( servo_master_message m ) {
                                servo_master_group::extract( m ).match(
                                    [&]( servo_master_command_variant var ) {
                                            var << tuple_match(
                                                [&]( cmd_tag< POSITION >, bangle angle ) {
                                                        res = angle;
                                                },
                                                [&]( auto tag, auto... ) {
                                                        FAIL() << tag;
                                                } );
                                    },
                                    [&]( auto ) {
                                            FAIL();
                                    } );
                        } );
                return res;
        }
};

TEST_P( servo_text_fixture, activate )
{

        ASSERT_TRUE( servo.has_task() );
        tick_n( 40 );
        ASSERT_FALSE( servo.has_task() );
}

TEST_P( servo_text_fixture, ctl_packet )
{
        // boot
        tick_n( 20 );

        master_servo_command_variant var =
            std::make_tuple( cmd_tag< GOTO >(), config->input_low, btime::from_float( 2.5f ) );
        servo.on_servo_packet( ctl_id{ 0 }, var )
            .match(
                [&]( servo_event ) {
                        SUCCEED();
                },
                [&]( servo_master_message m ) {
                        FAIL() << "Got message instead of event: " << m;
                } );

        var = std::make_tuple( cmd_tag< VOLTAGE >{} );
        servo.on_servo_packet( ctl_id{ 0 }, var )
            .match(
                [&]( servo_event ) {
                        FAIL() << "Got event instead of message";
                },
                [&]( servo_master_message ) {
                        SUCCEED();
                } );

        // TODO: bad message
}

TEST_P( servo_text_fixture, goto_event )
{
        tick_n( 20 );

        // move to def. position
        event_cmd< GOTO >( config->input_low, btime::from_float( 1.f ) );
        tick_n( 100, true, clk_time{ 10 } );
        EXPECT_EQ( get_pos(), config->input_low );

        event_cmd< GOTO >( config->input_high, btime::from_float( 4.f ) );
        tick_n( 100, true, clk_time{ 10 } );
        EXPECT_EQ( get_pos(), config->input_high );
}

TEST_P( servo_text_fixture, goto_reverse_event )
{
        tick_n( 20, true );

        servo_events_variant var = {
            lewan_goto_event{ lewan_angle::get< 0 >(), lewan_time::get< 4000 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 300, true, clk_time{ 10 } );

        EXPECT_EQ( get_pos(), bangle::from_float( 3.f ) );
}

TEST_P( servo_text_fixture, goto_out_of_deadzone )
{
        tick_n( 20, true );

        servo_events_variant var = {
            lewan_goto_event{ lewan_angle::get< 0 >(), lewan_time::get< 4000 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 4000 );

        EXPECT_EQ( get_pos(), bangle::from_float( 0 ) );
}

TEST_P( servo_text_fixture, spin )
{
        // note: remeber that lewan_time is in milliseconds, but the velocity is per second

        tick_n( 20 );

        servo_events_variant var = {
            lewan_spin_event{ lewan_angular_velocity::get< 100 >(), lewan_time::get< 500 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 500 );

        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );

        var = {
            lewan_spin_event{ lewan_angular_velocity::get< -50 >(), lewan_time::get< 500 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 500 );
        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );
}

TEST_P( servo_text_fixture, spin_until )
{
        tick_n( 20 );
        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );

        servo_events_variant var = { lewan_spin_until_event{
            lewan_angle::get< 500 >(), lewan_angular_velocity::get< 100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 5 * 110, true, clk_time{ 10 } );
        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );

        tick_n( 1000 );
        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );
}

TEST_P( servo_text_fixture, reverse_spin_until )
{
        tick_n( 20 );

        servo_events_variant var = {
            lewan_goto_event{ lewan_angle::get< 1000 >(), lewan_time::get< 5000 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 500, true, clk_time{ 10 } );

        var = { lewan_spin_until_event{ lewan_angle::get< 500 >(),
                                        lewan_angular_velocity::get< -100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 5 * 100, true, clk_time{ 10 } );
        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );
}

TEST_P( servo_text_fixture, spin_until_deadzone_pass )
{
        tick_n( 20 );
        tick_n( 20, true );

        servo_events_variant var = {
            lewan_goto_event{ lewan_angle::get< 100 >(), lewan_time::get< 100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 100, true, clk_time{ 10 } );
        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );

        var = { lewan_spin_until_event{ lewan_angle::get< 0 >(),
                                        lewan_angular_velocity::get< 100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 15 * 100, true, clk_time{ 10 } );

        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );
}

TEST_P( servo_text_fixture, reverse_spin_until_deadzone_pass )
{
        tick_n( 20 );

        servo_events_variant var = { lewan_spin_until_event{
            lewan_angle::get< 50 >(), lewan_angular_velocity::get< -100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 16 * 100, true, clk_time{ 10 } );

        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );
}

TEST_P( servo_text_fixture, spin_until_from_deadzone )
{
        tick_n( 20 );

        servo_events_variant var = { lewan_spin_until_event{
            lewan_angle::get< 250 >(), lewan_angular_velocity::get< 100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 16 * 1000, true );

        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );
}

TEST_P( servo_text_fixture, reverse_spin_until_from_deadzone )
{
        tick_n( 20 );

        servo_events_variant var = { lewan_spin_until_event{
            lewan_angle::get< 250 >(), lewan_angular_velocity::get< -100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );

        tick_n( 16 * 1000, true );

        EXPECT_EQ( get_pos(), bangle::from_float( 0.f ) );
}

TEST_P( servo_text_fixture, stuck_detector )
{
        tick_n( 20 );

        servo_events_variant var = { lewan_spin_until_event{
            lewan_angle::get< 250 >(), lewan_angular_velocity::get< 100 >() } };
        servo.on_event( ctl_id{ 0 }, this->clock_time, var, make_matcher() );
        for ( std::size_t i = 0; i < 500; i++ ) {
                this->clock_time += clk_time{ 10 };
                servo.tick( this->clock_time, this->make_matcher() );
        }

        EXPECT_FALSE( servo.has_task() );
}

INSTANTIATE_TEST_SUITE_P(
    servo_group,
    servo_text_fixture,
    testing::ValuesIn( generate_servo_params() ),
    []( const testing::TestParamInfo< servo_text_fixture::ParamType >& info ) {
            std::stringstream   ss;
            const servo_config& c = info.param;
            ss << *c.input_low << "_" << *c.input_high << "_" << c.servo_low << "_" << c.servo_high
               << "_" << c.angle_change_report << "_" << c.angle_stall_range;
            return ss.str();
    } );

// TEST ideas:
// - goto that crosses deadzone // how this should behave?

int main( int argc_in, char** argv_in )
{
        ros::Time::init();
        testing::InitGoogleTest( &argc_in, argv_in );

        return RUN_ALL_TESTS();
}
