#include "fw/serial.h"
#include "fw/servo/holder.h"
#include "node/lua_vm/def.h"
#include "node/lua_vm/extra_ser.h"

#include <ros/ros.h>
#include <schpin_lib/lua/repl.h>
#include <schpin_lib/lua/vm.h>
#include <schpin_lib_exp/config.h>

using namespace schpin;

// TODO:
//
//    make extension to protocol that allows json -> protocol serialization

struct msg_event
{
        nlohmann::json j;
};

struct msg_bind
{

        static constexpr std::string_view name = "msg";
        static constexpr std::string_view desc = "creates message out of json";

        inline static const auto arg =
            slua_struct< msg_event >( lua_key_arg< nlohmann::json >{ "json", "data" } );
        inline static const no_return_type res{};
};

struct dump_event
{
};

struct dump_bind
{
        static constexpr std::string_view name = "dump";
        static constexpr std::string_view desc = "dump information about servo";

        inline static const auto           arg = slua_struct< dump_event >();
        inline static const no_return_type res{};
};

struct quit_event
{
};

struct quit_bind
{
        static constexpr std::string_view name = "quit";
        static constexpr std::string_view desc = "closes the app";

        inline static const auto           arg = slua_struct< quit_event >();
        inline static const no_return_type res{};
};

struct slua_runner_def
{
        static auto get_def()
        {
                return std::make_tuple( slua_callback< msg_bind >{},
                                        slua_callback< dump_bind >{},
                                        slua_callback< quit_bind >() );
        }
};

template < auto ID >
struct nlohmann::adl_serializer< cmd_tag< ID > >
{
        static void to_json( nlohmann::json& j, cmd_tag< ID > )
        {
                j = magic_enum::enum_name( ID );
        }
        static void from_json( const nlohmann::json& j, cmd_tag< ID >& id )
        {

                auto opt_val = magic_enum::enum_cast< decltype( ID ) >( j.get< std::string >() );
                if ( !opt_val.has_value() ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "Failed to parse ID" );
                }

                if ( opt_val.value() != ID ) {
                        throw nlohmann::json::parse_error::create( -1, 0, "ID does not match" );
                }

                id = cmd_tag< ID >{};
        }
};

template < typename >
struct variant_serializer;

template < typename... Ts >
struct variant_serializer< std::variant< Ts... > >
{
        template < typename T >
        struct tag
        {
                using value_type = T;
        };

        static void to_json( nlohmann::json& j, const std::variant< Ts... >& var )
        {
                std::visit(
                    [&]( const auto& item ) {
                            j = item;
                    },
                    var );
        }

        static std::variant< Ts... > from_json( const nlohmann::json& j )
        {
                std::optional< std::variant< Ts... > > opt_res;

                auto f = [&]( auto tagv ) {
                        using T = typename decltype( tagv )::value_type;
                        try {
                                opt_res = std::variant< Ts... >{ j.get< T >() };
                                return true;
                        }
                        catch ( ... ) {
                                return false;
                        }
                };

                ( f( tag< Ts >{} ) || ... || false );
                if ( !opt_res ) {
                        throw nlohmann::json::parse_error::create(
                            -1, 0, "no variant alternative can be parsed" );
                }
                return *opt_res;
        }
};

template <>
struct nlohmann::adl_serializer< master_servo_command_variant >
  : variant_serializer< master_servo_command_variant >
{
};

template <>
struct nlohmann::adl_serializer< master_servo_config_variant >
  : variant_serializer< master_servo_config_variant >
{
};

template <>
struct nlohmann::adl_serializer< typename prot_config_types< servo_config >::parameters_variant >
  : variant_serializer< typename prot_config_types< servo_config >::parameters_variant >
{
};

struct servo_tester_config
{
        servo_type            type;
        servo_id              id;
        std::filesystem::path dev;

        static auto ros_param_get()
        {
                return std::make_tuple(  //
                    ros_param< servo_type >( "stype", "Servo type" ),
                    ros_param< servo_id >( "sid", "Servo id" ),
                    ros_param< std::filesystem::path >( "device",
                                                        "Device to open for connection" ) );
        }
};

class servo_runner
{
        serial_connection          ser_;
        servo_holder               servo_;
        slua_vm< slua_runner_def > vm_;
        slua_repl                  repl_;

        public:
        servo_runner( const servo_tester_config& conf, serial_connection ser )
          : ser_( ser )
          , servo_( *servo_holder::make( conf.type, conf.id, get_system_time() ) )
        {
        }

        // TODO: duplication from control_pc :/
        auto make_matcher()
        {
                return match(
                    []( control_tag, servo_id, auto ) {},
                    []( activate_tag, servo_id id ) {
                            SCHPIN_INFO_LOG( "lewan_servo", "Got activation of: " << id );
                    },
                    [this]( servo_tag, s_ctl_id msg_id, reply r, const master_lewan_message& m ) {
                            ser_.write( m );
                            if ( r == reply::NO ) {
                                    return;
                            }
                            receive_serial_msg( msg_id );
                    },
                    [this]( servo_tag, s_ctl_id msg_id, reply, const master_dynmx_message& m ) {
                            SCHPIN_INFO_LOG( "servo_runner", "TO DYN: " << std::hex << m );
                            ser_.write( m );
                            ser_.read< dynmx_master_message::max_size >()
                                .bind_left( [&]( const auto& data ) {
                                        SCHPIN_INFO_LOG(
                                            "servo_runner",
                                            "FR DYN: " << std::hex << convert_view< int >( data ) );
                                        return dynmx_master_protocol::extract( data ).convert_right(
                                            [&]( const protocol_error_record rec ) {
                                                    return E( protocol_error ) << rec;
                                            } );
                                } )
                                .match(
                                    em::uncurry |
                                        [&]( dynmx_header, dynmx_servo_id, auto tpl, dynmx_crc ) {
                                                const dynmx_response_error& err =
                                                    std::get< 1 >( tpl );
                                                if ( err != dynmx_no_error ) {
                                                        SCHPIN_ERROR_LOG(
                                                            "runner",
                                                            "Dynmx return error: " << err );
                                                }
                                                const dynmx_buffer& buffer = std::get< 2 >( tpl );

                                                servo_.on_servo_receive(
                                                    get_system_time(), msg_id, buffer );
                                        },
                                    [&]( auto err ) {
                                            SCHPIN_ERROR_LOG( "servo_runner", err );
                                    } );
                    } );
        }
        void receive_serial_msg( s_ctl_id msg_id )
        {
                static constexpr std::size_t n = lewan_master_message::max_size;
                ser_.read< n >().match(
                    [&]( const em::static_vector< uint8_t, n >& data ) {
                            lewan_sequencer< lewan_master_message > sequencer;

                            bool load_fired = sequencer.load( data, [&]( lewan_master_message m ) {
                                    parse_message( msg_id, m );
                            } );

                            if ( !load_fired ) {
                                    SCHPIN_ERROR_LOG(
                                        "control",
                                        "servo not replying, got:" << convert_view< int >( data ) );
                            }
                    },
                    [&]( auto err ) {
                            SCHPIN_ERROR_LOG( "control", err );
                    } );
        }
        void parse_message( s_ctl_id msg_id, const lewan_master_message& m )
        {
                lewan_master_protocol::extract( m )
                    .bind_left( em::uncurry |
                                [&]( lewan_servo_id, lewan_master_submessage m ) {
                                        return lewan_master_group::extract( m );
                                } )
                    .match(
                        [&]( lewan_master_variant msg_var ) {
                                servo_.on_servo_receive( get_system_time(), msg_id, msg_var );
                        },
                        [&]( auto err ) {
                                SCHPIN_ERROR_LOG( "lewan_servo", err );
                        } );
        }

        void tick()
        {
                servo_.tick( get_system_time(), make_matcher() );
                servo_.poll( get_system_time(), 0, make_matcher() );
                std::string msg = repl_.prompt();
                vm_.load( msg ).optionally_log();
                vm_.full_exec( [&]( lua_State*, auto event ) {
                           handle_event( event );
                   } )
                    .optionally_log();
        }

        void handle_event( quit_event )
        {
                std::exit( 0 );
        }

        void handle_event( dump_event )
        {
                servo_.dump_info();
        }

        void handle_event( msg_event e )
        {
                auto var = e.j.get< master_servo_command_variant >();

                servo_.on_servo_packet( ctl_id{ 0 }, var )
                    .match(
                        [&]( servo_event se ) {
                                servo_.on_event(
                                    ctl_id{ 0 }, get_system_time(), se.sub_event, make_matcher() );
                        },
                        [&]( servo_master_message m ) {
                                // TODO: this should return as json
                                SCHPIN_INFO_LOG( "servo_runner", m );
                        } );
        }
};

template <>
struct schpin::ros_param_serializer< servo_type > : schpin::ros_param_serializer_enum< servo_type >
{
};

int main( int argc, char* argv[] )
{

        ros::init( argc, argv, "servo_tester" );
        ros::AsyncSpinner spinner{ 2 };
        spinner.start();

        get_config< servo_tester_config >( "~" )
            .bind_left( [&]( const servo_tester_config& conf ) {
                    return serial_connection::make( conf.dev )
                        .convert_left( [&]( serial_connection c ) {
                                return std::make_tuple( conf, c );
                        } );
            } )
            .match(
                em::uncurry |
                    [&]( servo_tester_config conf, serial_connection p ) {
                            servo_runner runnner{ conf, p };

                            while ( true ) {
                                    runnner.tick();
                            }
                    },
                [&]( auto e ) {
                        std::cerr << e;
                } );
}

