#include "schpin_lib/algorithm.h"
#include "schpin_lib/error.h"
#include "schpin_lib/print.h"
#include "schpin_lib/quantity.h"
#include "schpin_lib_exp/algorithm.h"
#include "schpin_lib_exp/bounded.h"

#include <filesystem>
#include <optional>
#include <ros/ros.h>

#pragma once

namespace schpin
{

using config_error = error< struct config_errorTag >;

inline std::ostream& operator<<( std::ostream& os, const config_error& e )
{
        return os << simplified_error{ e };
}

template < typename T >
struct ros_param_serializer
{
        static em::either< T, config_error > load( const std::string& prefix )
        {
                auto arg_tuple = T::ros_param_get();

                return std::apply(
                    [&]( auto... item ) {
                            return assemble_left_collect_right( item.load( prefix )... )
                                .convert_left( [&]( auto value_tpl ) {
                                        return std::apply(
                                            [&]( auto... values ) {  //
                                                    return T{ values... };
                                            },
                                            value_tpl );
                                } )
                                .convert_right( [&]( const auto& errors ) {
                                        return E( config_error ).group_attach( errors );
                                } );
                    },
                    arg_tuple );
        }
};

template < int Len, int Mass, int Time, int Current, int Temp, int Mol, int Li, int Angle, int Byte >
struct ros_param_serializer<
    emlabcpp::physical_quantity< Len, Mass, Time, Current, Temp, Mol, Li, Angle, Byte > >
  : ros_param_serializer< emlabcpp::quantity<
        emlabcpp::physical_quantity< Len, Mass, Time, Current, Temp, Mol, Li, Angle, Byte >,
        float > >
{
};

template < typename T, typename ValueType >
struct ros_param_serializer< em::quantity< T, ValueType > >
{
        static em::either< T, config_error > load( const std::string& prefix )
        {
                return ros_param_serializer< ValueType >::load( prefix )
                    .template construct_left< T >();
        }
};

template < typename T, T FromVal, T ToVal >
struct ros_param_serializer< bounded< T, FromVal, ToVal > >
{
        static em::either< bounded< T, FromVal, ToVal >, config_error >
        load( const std::string& prefix )
        {
                return ros_param_serializer< T >::load( prefix ).bind_left(
                    [&]( T val ) -> em::either< bounded< T, FromVal, ToVal >, config_error > {
                            auto opt_val = bounded< T, FromVal, ToVal >::make( val );
                            if ( opt_val ) {
                                    return { *opt_val };
                            }
                            return { E( config_error ) << "Bounded value out of range" };
                    } );
        }
};

template < typename T >
struct ros_param_serializer< std::map< std::string, T > >
{
        static em::either< std::map< std::string, T >, config_error >
        load( const std::string& prefix )
        {
                XmlRpc::XmlRpcValue key_list;
                if ( !ros::param::get( prefix, key_list ) ) {
                        return { E( config_error ) << "Failed  to load keys" };
                }

                auto eithers = map_f_to_v( key_list, [&]( const auto& item ) {
                        std::string key = item.first;
                        return ros_param_serializer< T >::load( prefix + "/" + key )
                            .convert_left( [&]( T item ) {
                                    return std::make_pair( key, item );
                            } );
                } );

                auto parted = partition_eithers( eithers );
                if ( !parted.right.empty() ) {
                        return { E( config_error ).group_attach( parted.right ) << "Failed to load "
                                                                                   "map" };
                }

                std::map< std::string, T > map;
                for ( auto pair : parted.left ) {
                        map.insert( pair );
                }

                return { map };
        }
};

template < typename OutputType, typename SerializationType = OutputType >
struct ros_param_serializer_basic
{
        static em::either< OutputType, config_error > load( const std::string& prefix )
        {
                SerializationType item;
                if ( ros::param::get( prefix, item ) ) {
                        return { static_cast< OutputType >( item ) };
                }
                return { E( config_error ) };
        }
};

template <>
struct ros_param_serializer< std::string > : ros_param_serializer_basic< std::string >
{
};
template <>
struct ros_param_serializer< double > : ros_param_serializer_basic< double >
{
};
template <>
struct ros_param_serializer< float > : ros_param_serializer_basic< float >
{
};
template <>
struct ros_param_serializer< uint8_t > : ros_param_serializer_basic< uint8_t, int >
{
};
template <>
struct ros_param_serializer< int8_t > : ros_param_serializer_basic< int8_t, int >
{
};
template <>
struct ros_param_serializer< uint16_t > : ros_param_serializer_basic< uint16_t, int >
{
};
template <>
struct ros_param_serializer< int16_t > : ros_param_serializer_basic< int16_t, int >
{
};
template <>
struct ros_param_serializer< uint32_t > : ros_param_serializer_basic< uint32_t, int >
{
};
template <>
struct ros_param_serializer< int32_t > : ros_param_serializer_basic< int32_t, int >
{
};

template <>
struct ros_param_serializer< std::filesystem::path >
{
        static em::either< std::filesystem::path, config_error > load( const std::string& prefix )
        {
                return ros_param_serializer< std::string >::load( prefix )
                    .template construct_left< std::filesystem::path >();
        }
};

template < typename T >
struct ros_param_serializer< std::optional< T > >
{
        static em::either< std::optional< T >, config_error > load( const std::string& prefix )
        {
                return ros_param_serializer< T >::load( prefix )
                    .convert_left( [&]( T val ) {  //
                            return std::make_optional< T >( val );
                    } )
                    .bind_right( [&]( auto ) -> em::either< std::optional< T >, config_error > {
                            return { std::optional< T >{} };
                    } );
        }
};

template < typename T >
struct ros_param_serializer_enum
{
        static em::either< T, config_error > load( const std::string& prefix )
        {
                return ros_param_serializer< std::string >::load( prefix ).bind_left(
                    [&]( std::string val ) -> em::either< T, config_error > {
                            auto eval = magic_enum::enum_cast< T >( val );
                            if ( !eval.has_value() ) {
                                    return E( config_error )
                                               .group_attach( magic_enum::enum_names< T >() )
                                           << "String " << val
                                           << " is not valid enum value, see list for valid values:";
                            }
                            return eval.value();
                    } );
        }
};

template < typename T >
class ros_param
{
        std::string parameter_name_;
        std::string description_;

        public:
        ros_param( std::string parameter_name, std::string description )
          : parameter_name_( std::move( parameter_name ) )
          , description_( std::move( description ) )
        {
        }

        em::either< T, config_error > load( const std::string& prefix = "" )
        {
                return ros_param_serializer< T >::load( prefix + "/" + parameter_name_ )
                    .convert_right( [&]( config_error e ) -> config_error {
                            return E( config_error ).attach( e )
                                   << "Failed to get config parameter: " << term_color::MAGENTA
                                   << prefix + parameter_name_ << term_color::RESET
                                   << " desc: " << description_;
                    } );
        }
};

template < typename T >
class ros_opt_param
{
        std::string name_;
        std::string description_;
        T           default_;

        public:
        ros_opt_param( std::string name, std::string description, T def )
          : name_( std::move( name ) )
          , description_( std::move( description ) )
          , default_( def )
        {
        }

        em::either< T, config_error > load( const std::string& prefix = "" )
        {
                return ros_param_serializer< T >::load( prefix + "/" + name_ )
                    .bind_right( [&]( const config_error& e ) -> em::either< T, config_error > {  //
                            SCHPIN_INFO_LOG( "config",
                                             "Parameter " << name_
                                                          << " was not loaded due to: " << e );
                            return default_;
                    } );
        }
};

template < typename Config >
em::either< Config, config_error > get_config( const std::string& prefix = "" )
{
        return ros_param_serializer< Config >::load( prefix );
}

}  // namespace schpin
