#include "emlabcpp/zip.h"
#include "schpin_lib_exp/convert_iterator.h"
#include "schpin_lib_exp/overload.h"
#include "schpin_lib_exp/protocol/item.h"
#include "schpin_lib_exp/static_view.h"

#pragma once

namespace schpin
{

template < typename MsgCounterType = uint8_t, typename... Items >
struct protocol_tuple
{

        template < typename NewMsgCounterType >
        using with_msg_counter = protocol_tuple< NewMsgCounterType, Items... >;

        template < typename... NewItems >
        using with_items = protocol_tuple< MsgCounterType, Items..., NewItems... >;

        template < typename... NewItems >
        using with_promoted_items =
            protocol_tuple< MsgCounterType, Items..., protocol_recursive_promoter< NewItems >... >;

        using item_tuple  = std::tuple< Items... >;
        using input_type  = typename protocol_item< item_tuple >::input_type;
        using output_type = typename protocol_item< item_tuple >::output_type;

        static constexpr uint16_t    item_n = static_cast< uint16_t >( sizeof...( Items ) );
        static constexpr std::size_t size   = ( protocol_item< Items >::size + ... + 0 );
        using message_t                     = prot_message< MsgCounterType, size, protocol_tuple >;

        static std::string stringify( const input_type& v )
        {
                std::stringstream ss;
                em::for_each(
                    em::make_sequence_tuple_t< sizeof...( Items ) >{}, [&]( auto integ_const ) {
                            constexpr std::size_t i = decltype( integ_const )::value;
                            using t                 = std::variant_alternative_t< i, item_tuple >;

                            ss << protocol_item< t >::stringify( std::get< i >( v ) ) << " | ";
                    } );

                return ss.str();
        }

        static std::string stringify( const message_t& m )
        {
                std::stringstream ss;
                extract( m ).match(
                    [&]( const input_type& v ) {
                            ss << "{ " << stringify( v ) << " || " << convert_view< int >( m )
                               << " }";
                    },
                    [&]( auto ) {
                            ss << convert_view< int >( m );
                    } );
                return ss.str();
        }

        template < typename... Args >
        static constexpr output_type make_val( Args&&... args )
        {
                return output_type{ args... };
        }

        template < typename... Args >
        static constexpr message_t serialize( Args&&... items_val )
        {
                std::array< uint8_t, size > buff = { 0 };
                std::size_t used = serialize_at( make_static_view( buff ), items_val... );
                return message_t{ buff, used };
        }

        template < typename... Args >
        static constexpr std::size_t serialize_at( static_view< size, uint8_t* > target,
                                                   Args&&... items_val )
        {
                return protocol_item< item_tuple >::serialize_at( target,
                                                                  input_type{ items_val... } );
        }

        template < typename Message, typename... Visitors >
        static std::size_t match( Message&& m, Visitors&&... vis )
        {
                return match( m, schpin::match( vis... ) );
        }

        template < typename Message, typename Visitor >
        static std::size_t match( Message&& m, Visitor&& vis )
        {
                return protocol_item< item_tuple >::deserialize(
                    m,
                    schpin::match(
                        [&]( value_tag, output_type tpl ) {
                                std::apply(
                                    [&]( auto... items ) {
                                            vis( items... );
                                    },
                                    tpl );
                        },
                        [&]( error_tag, protocol_error_record rec ) {
                                vis( error_tag{}, rec );
                        } ) );
        }

        template < typename Message >
        static em::either< output_type, protocol_error_record > extract( Message&& m )
        {
                // TODO: the error callback can be called multiple times, maybe allow record notice
                // that?
                em::either< output_type, protocol_error_record > res = protocol_error_record{};
                match(
                    m,
                    [&]( error_tag, protocol_error_record rec ) {
                            res = rec;
                    },
                    [&]( auto... items ) {
                            res = std::make_tuple( items... );
                    } );
                return res;
        }
};

}  // namespace schpin
