#include "lua_vm.h"

#include <schpin_lib/lua/print.h>
#include <schpin_lib/lua/vm.h>

using namespace schpin;

struct slua_motion_vm_def
{
        static auto get_def()
        {
                return std::make_tuple(
                    slua_proc(  //
                        "load_environment",
                        "Loads the model of the environment",
                        slua_struct< load_environment_event >(  //
                            lua_key_arg< std::string >{ "filename",
                                                        "path to the environment file" } ) ),
                    slua_proc(  //
                        "set_goal",
                        "Sets the goal positions for the planner",
                        slua_struct< set_goal_pose_event >( slua_pose_args(  //
                            "point",
                            "orientation",
                            "Sets the target pose for movement, specify either point, orientation "
                            "or "
                            "both" ) ) ),
                    slua_proc(  //
                        "set_current",
                        "Sets the current position of the robot in the environment manually",
                        slua_struct< set_current_pose_event >( slua_pose_args(  //
                            "point",
                            "orientation",
                            "Sets the current pose for movement, specify either point, orientation "
                            "or "
                            "both" ) ) ),
                    slua_proc( "plan", "Starts the planning process", slua_struct< plan_event >() ),
                    slua_proc( "clear", "Clears the data", slua_struct< clear_event >() ),
                    slua_proc(  //
                        "export",
                        "Exports the movement plan to .lua file as appropiate commands",
                        slua_struct< export_event >(  //
                            lua_key_arg< std::string >{ "filename",
                                                        "Path to the exported file" } ) ),
                    slua_proc(  //
                        "show_body_collision",
                        "Shows collisio nstructure used for body",
                        slua_struct< show_body_collision_event >() ),
                    slua_proc(  //
                        "show_leg_graph",
                        "Shows movement graph of an leg",
                        slua_struct< show_leg_graph_event >(
                            lua_opt_key_arg< int >{ "index", "index of an leg", -1 } ) ),
                    slua_proc(  //
                        "show_workarea",
                        "Shows workarea of leg",
                        slua_struct< show_workarea_event >(
                            lua_opt_key_arg< int >{ "index", "index of an leg", -1 } ) ),
                    slua_proc(  //
                        "print_robot_model",
                        "Prints various information about the robot",
                        slua_struct< print_model_event >() ),
                    slua_proc(  //
                        "print_pose_cache",
                        "Prints cache of pose informations",
                        slua_struct< print_pose_cache_event >() ),
                    slua_proc(  //
                        "make_body_vertexes",
                        "Create vertexes in graph for body movement",
                        slua_struct< make_vertexes_event >(
                            lua_key_arg< unsigned >{ "count", "Number of vertexes" },
                            lua_opt_key_arg< unsigned >{ "v", "Verbosity level", 0 } ) ),
                    slua_proc(  //
                        "find_body_path",
                        "Finds path in existing body graph",
                        slua_struct< find_body_path_event >() ),
                    slua_proc(  //
                        "make_body_until_path",
                        "Creates vertexes until path is found",
                        slua_struct< make_until_path_event >() ),
                    slua_proc(  //
                        "export_path",
                        "Exports the movement path to .lua file as appropiate commands",
                        slua_struct< export_path_event >(  //
                            lua_key_arg< std::string >{ "filename",
                                                        "Path to the exported file" } ) ),
                    slua_proc(  //
                        "quit",
                        "Stops the motion planner",
                        slua_struct< quit_event >() ) );
        }
};

struct lua_vm_eventmaker::impl
{
        slua_vm< slua_motion_vm_def > vm;

        opt_error< slua_error > load_file( std::filesystem::path p )
        {
                return vm.load( p );
        }

        bool is_finished()
        {
                return vm.load_finished();
        }

        em::either< lua_res, slua_error > tick( std::function< void( lua_event_var ) > event_cb )
        {
                return vm.exec( [&]( lua_State*, auto var ) { event_cb( var ); } )
                       << match( [&]( const auto& item ) {
                                  return em::either< lua_res, slua_error >{ item };
                          } );
        }
};

lua_vm_eventmaker::lua_vm_eventmaker()
  : impl_ptr_{ std::make_unique< lua_vm_eventmaker::impl >() }
{
}

bool lua_vm_eventmaker::is_finished()
{
        return impl_ptr_->is_finished();
}

opt_error< slua_error > lua_vm_eventmaker::load_file( std::filesystem::path p )
{
        return impl_ptr_->load_file( p );
}

em::either< lua_res, slua_error >
lua_vm_eventmaker::tick( std::function< void( lua_event_var ) > fun )
{
        return impl_ptr_->tick( fun );
}

lua_vm_eventmaker::~lua_vm_eventmaker() = default;

